/*
fluxy. A Realtime 2d Path Tracer
---
Copyright (C) 2008  Sebastian Mach (*1983)
seb@greenhybrid.net | http://greenhybrid.net
---
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef __cplusplus
#include <cstdlib>
#else
#include <stdlib.h>
#endif
#ifdef __APPLE__
#include <SDL/SDL.h>
#else
#include <SDL/SDL.h>
#endif


#include <time.h>

#include "colorfrgb.h"
#include "surface.h"
#include "ptintegrator.h"
#include "linelist.h"
#include "raytracer.h"


template<class t_surface>
void
smooth(
    t_surface &target,
    const t_surface &source
)
{
    const int height = target.getHeight();
    const int width  = target.getWidth();

    for( int y=0; y<height; ++y ){
        for( int x=0; x<width; ++x ){
            target(x,y) = ColorFRGB(0.0,0.0,0.0);
        }
    }

    for( int y=0; y<height; ++y ){
        for( int x=0; x<width; ++x ){
            //target( x, y ) = source(x,y)*0.333 + source(x-1,y)*0.333 + source(x+1,y)*0.333;
            ColorFRGB col = source(x,y);
            float r,g,b;
            col.to_rgb(r,g,b);
            float I = 0.299*r + 0.584*g + 0.144*b;
            float deploy = 0.01;
            ColorFRGB dist( r * (deploy/I), g * (deploy/I), b * (deploy/I) );
            int u=x,v=y;
            while( I > 0.0f ){
                switch( rand()%4 ){
                    case 0: ++u; break;
                    case 1: --u; break;
                    case 2: ++v; break;
                    case 3: --v; break;
                };
                if( u>=0 && u<width && v>=0 && v<height )
                    target(u,v) += dist;
                I -= deploy;
            }
        }
    }
}

template<class t_surface>
void
draw(
	SDL_Surface *p_target,
	const t_surface &surface
)
{
	if( SDL_MUSTLOCK(p_target) && SDL_LockSurface(p_target)<0 )
		return;
	int x,y;
	for( y=0; y<p_target->h; y++ ){
	    // TODO FIX we are currently assuming a 32bit SDL buffer
	    // TODO get rid of pointer arithmetic
		Uint32 *bufp   = (Uint32*)p_target->pixels + y*(p_target->pitch>>2);
		for( x=0; x<p_target->w; x++ ){
		    float r,g,b;
			((ColorFRGB)surface(x,y)).to_rgb( r, g, b );
            r = r<0?0:r>1?1:r;
            g = g<0?0:g>1?1:g;
            b = b<0?0:b>1?1:b;
			*(bufp++) =
				SDL_MapRGB(p_target->format,
					(unsigned char)(255.0*r),
					(unsigned char)(255.0*g),
					(unsigned char)(255.0*b)
				);
		}
	}

	if( SDL_MUSTLOCK(p_target) )
		SDL_UnlockSurface(p_target);
	SDL_Flip(p_target);
}


int main ( int argc, char** argv ) {
    // initialize SDL video
    if ( SDL_Init( SDL_INIT_VIDEO ) < 0 ) {
        printf( "Unable to init SDL: %s\n", SDL_GetError() );
        return 1;
    }

    const unsigned int width = 256, height = width;

    typedef RayTracer<LineList,PTIntegrator> raytracer_t;
    raytracer_t rayTracer;
    raytracer_t::scene_t &scene = rayTracer.scene;
    Surface<ColorFRGB> surface(width, height), smoothedSurface(width,height), tmpSurface(width,height);

    // make sure SDL cleans up before exit
    atexit(SDL_Quit);

    // create a new window
    SDL_Surface* screen = SDL_SetVideoMode(width, height, 32,
                                           SDL_HWSURFACE|SDL_DOUBLEBUF);
    if ( !screen ) {
        printf("Unable to set %ux%u video: %s\n", width, height, SDL_GetError());
        return 1;
    }

    // load an image
    /*SDL_Surface* bmp = SDL_LoadBMP("cb.bmp");
    if (!bmp) {
        printf("Unable to load bitmap: %s\n", SDL_GetError());
        return 1;
    }*/

    // fps
    unsigned int fpsNumFrames = 0;
    float fpsFPS = 0.0f;
    clock_t fpsLastMeasure = clock();
    float lumiX=0.5, lumiY=0.5;
    // program main loop
    bool done = false;
    //int f = 0;
    while (!done /*&& ++f<100*/ ) {

        //const float ftime = static_cast<float>(clock()) / static_cast<float>(CLOCKS_PER_SEC);
        scene.beginSceneDefinition();{
            /*
            float s = 0.49;
            scene.addLine( Line( Vector(-s,-s), Vector(-s,+s), Material(ColorFRGB(0.5,0.5,0.5),false) ) );
            scene.addLine( Line( Vector(-s,+s), Vector(+s,+s), Material(ColorFRGB(1.0,0.0,0.0),false) ) );
            scene.addLine( Line( Vector(+s,+s), Vector(+s,-s), Material(ColorFRGB(0.5,0.5,0.5),false) ) );
            scene.addLine( Line( Vector(+s,-s), Vector(-s,-s), Material(ColorFRGB(0.0,1.0,0.0),false) ) );
            */

            scene.addLine( Line( Vector(-0.124,-0.2), Vector(0.124,-0.2), Material(ColorFRGB(1.0,0.25,0.25)) ) );
            scene.addLine( Line( Vector(-0.5,-0.4), Vector(0.5,-0.4), Material(ColorFRGB(0.25,1.0,0.25)) ) );

            scene.addLine(
                Line( Vector(lumiX-0.1,lumiY), Vector(lumiX+0.1,lumiY), Material(ColorFRGB(4.0,4.0,4.0),true) )
            );
            /*scene.addLine(
                Line( Vector(-0.125+0.2*sinf(0.7*ftime*0.25),0.20+0.15*sinf(ftime*0.25)), Vector(0.125+0.2*sinf(0.7*ftime*0.25),0.20), Material(ColorFRGB(2.0,2.0,2.0),true) )
            );*/
        }scene.endSceneDefinition();
        if( clock() - fpsLastMeasure > CLOCKS_PER_SEC ){
            fpsLastMeasure = clock();
            fpsFPS = static_cast<float>(fpsNumFrames);
            fpsNumFrames = 0;
            char caption[256];
            sprintf( caption, "fps:%.2f", fpsFPS );
            SDL_WM_SetCaption( caption, "..." );
        }

        // message processing loop
        SDL_Event event;
        while (SDL_PollEvent(&event)) {
            // check for messages
            switch (event.type) {
                // exit if the window is closed
                case SDL_QUIT:
                    done = true;
                    break;

                // check for keypresses
                case SDL_KEYDOWN: {
                    // exit if ESCAPE is pressed
                    if (event.key.keysym.sym == SDLK_ESCAPE)
                        done = true;
                    break;
                }
                // check for keyups
                case SDL_KEYUP: {
                    // exit if ESCAPE is pressed
                    if (event.key.keysym.sym == SDLK_ESCAPE)
                        done = true;
                    break;
                }


                case SDL_MOUSEMOTION: {
                    Uint8 keystate = SDL_GetMouseState(NULL, NULL);

                    if( keystate & SDL_BUTTON(1) ){
                        lumiX += 0.01*static_cast<float>(event.motion.xrel);
                        lumiY -= 0.01*static_cast<float>(event.motion.yrel);
                    }
                } break;
            } // end switch
        } // end of message processing

        // DRAWING STARTS HERE

        // clear screen
        //SDL_FillRect(screen, 0, SDL_MapRGB(screen->format, 0, 0, 0));

        // draw bitmap
        //SDL_BlitSurface(bmp, 0, screen, &dstrect);

        rayTracer.trace( surface );
        /*smooth( smoothedSurface, surface );
        draw( screen, smoothedSurface );*/
        draw( screen, surface );
        ++fpsNumFrames;
        // DRAWING ENDS HERE

        // finally, update the screen :)
        SDL_Flip(screen);
    } // end main loop

    // free loaded bitmap
    //SDL_FreeSurface(bmp);

    // all is well ;)
    printf("Exited cleanly\n");
    return 0;
}
